//*****************************************************************************
// USB read/write utility for FTDI FT245 series.
// Tested on Cygwin g++
//*****************************************************************************

#include <stdio.h>
#include <stdlib.h>
#include <w32api/windows.h>
#include <time.h>
#include "ftd2xx.h"

#define N_ELEMS(A) (sizeof(A)/sizeof((A)[0]))

//*****************************************************************************
// Declare USB procedures. This part is application independent.
//*****************************************************************************

static const char* getFtErrorMessage(FT_STATUS ftStatus);

// Enumerate procedure informations as triples
// (Procedure name, param types, param names). 

#define USB_PROCS \
  \
PROC(FT_Open, \
     (int iDevice, \
      FT_HANDLE *pftHandle), \
     (iDevice, pftHandle)) \
PROC(FT_ListDevices, \
     (PVOID pvArg1, \
      PVOID pvArg2, \
      DWORD dwFlags), \
     (pvArg1, pvArg2, dwFlags)) \
PROC(FT_Close, \
     (FT_HANDLE ftHandle), \
     (ftHandle)) \
PROC(FT_Read, \
     (FT_HANDLE ftHandle, \
      LPVOID lpBuffer, \
      DWORD dwBytesToRead, \
      LPDWORD lpdwBytesReturned), \
     (ftHandle, lpBuffer, dwBytesToRead, lpdwBytesReturned)) \
PROC(FT_Write, \
     (FT_HANDLE ftHandle, \
      LPVOID lpBuffer, \
      DWORD dwBytesToWrite, \
      LPDWORD lpdwBytesWritten), \
     (ftHandle, lpBuffer, dwBytesToWrite, lpdwBytesWritten)) \
PROC(FT_ResetDevice, \
     (FT_HANDLE ftHandle), \
     (ftHandle)) \
PROC(FT_SetTimeouts, \
     (FT_HANDLE ftHandle, \
      DWORD dwReadTimeout, \
      DWORD dwWriteTimeout), \
     (ftHandle, dwReadTimeout, dwWriteTimeout))

// enumerate procedure IDs
#define PROC(NAME, PARAMTYPES, PARAMLIST) ID_##NAME, 
enum { USB_PROCS FT_MAX };
#undef PROC


static void *ftFunctions[FT_MAX]; // pointers to USB procedures
static HMODULE usbDll = NULL;     // handle of USB DLL


// enumerate procedure names
#define PROC(NAME, PARAMTYPES, PARAMLIST) #NAME ,
static const char *usbProcNames[] = { USB_PROCS };
#undef PROC

// declare procedure types
#define PROC(NAME, PARAMTYPES, PARAMLIST) \
  typedef FTD2XX_API FT_STATUS (WINAPI *TYPE_##NAME) PARAMTYPES;
USB_PROCS
#undef PROC


//generate wrapper function of each procedure
#define PROC(NAME, PARAMTYPES, PARAMLIST) \
  FT_STATUS USB_##NAME PARAMTYPES { \
    return (*(TYPE_##NAME)(ftFunctions[ID_##NAME])) PARAMLIST; \
  }
USB_PROCS
#undef PROC

// generate each procedure's wrapper function with error check.
#define PROC(NAME, PARAMTYPES, PARAMLIST) \
  FT_STATUS EC_USB_##NAME PARAMTYPES { \
    FT_STATUS ftStatus = USB_##NAME PARAMLIST; \
    if(!FT_SUCCESS(ftStatus)){ \
      fprintf(stderr, "Error : " #NAME "failed : %s\n", \
              getFtErrorMessage(ftStatus)); \
      exit(1); \
    } \
    return ftStatus; \
  }
USB_PROCS
#undef PROC


//-----------------------------------------------------------------------------
// returns an error message coressponds to the ftStatus
//-----------------------------------------------------------------------------
static const char* getFtErrorMessage(FT_STATUS ftStatus)
{
  static const char* ftErrorMessages[] = {
    "No error",
    "Invalid Handle",
    "Device not found",
    "Device not opened",
    "I/O error",
    "Insufficient resouces",
    "Invalid parameter",
  };
  
  if(((UINT)ftStatus) >= N_ELEMS(ftErrorMessages)){
    return "Unknown Error";
  } else {
    return ftErrorMessages[ftStatus];
  } 
}

//-----------------------------------------------------------------------------
// setupDll
//-----------------------------------------------------------------------------
static void setupDll()
{
  unsigned u;
  const char* filename = "Ftd2xx.dll";
  usbDll = LoadLibrary(filename);
  if(usbDll == NULL) {
    fprintf(stderr, "Error: Can't Load %s\n", filename);
    exit(1);
  }
	
  for(u = 0; u < N_ELEMS(usbProcNames); u++){
    void *proc = (void*)GetProcAddress(usbDll, usbProcNames[u]);
    if(proc == NULL){
      fprintf(stderr, "Error : Can't Find %s\n", usbProcNames[u]);
      exit(1);
    }
    ftFunctions[u] = proc;
  }
}

//-----------------------------------------------------------------------------
// cleanupDll
//-----------------------------------------------------------------------------
static void cleanupDll()
{
  if(usbDll != NULL){
    FreeLibrary(usbDll);
    usbDll = NULL;
  }
}

//*****************************************************************************
// Applicationd dependent part.
//*****************************************************************************

const  FT_HANDLE FT_NULL_HANDLE = (FT_HANDLE)-1;
// static FT_HANDLE ftHandle = FT_NULL_HANDLE;
static FT_HANDLE ftHandle = (FT_HANDLE)-1;

//-----------------------------------------------------------------------------
// cleanApp
//-----------------------------------------------------------------------------
static void cleanApp()
{
  if(ftHandle != FT_NULL_HANDLE){
    USB_FT_ResetDevice(ftHandle);
    USB_FT_Close(ftHandle);
    ftHandle = FT_NULL_HANDLE;
  }
  cleanupDll();
}

//-----------------------------------------------------------------------------
// initUSB
//-----------------------------------------------------------------------------
static void initUSB()
{
  setupDll();
  atexit(cleanApp);
}

//-----------------------------------------------------------------------------
// main
//-----------------------------------------------------------------------------
int main(int argc, char* argv[])
{
  auto int i;
  auto char buffer[1024];
  auto long int count;
  auto long int tick;
  initUSB();
  EC_USB_FT_Open(0, &ftHandle);
  for (i = 1; i < argc; i++) {
    if (!strcmp(argv[i], "-t")) {
      if (++i >= argc) break;
      EC_USB_FT_SetTimeouts(ftHandle, atoi(argv[i]), atoi(argv[i]));
    }
  }
  EC_USB_FT_ResetDevice(ftHandle);
  tick = -GetTickCount();
  while (count = fread(buffer, 1, sizeof(buffer), stdin)) {
    EC_USB_FT_Write(ftHandle, buffer, count, &count);
  }
  while (EC_USB_FT_Read(ftHandle, buffer, sizeof(buffer), &count), count) {
    fprintf(stderr, "Message : %ld ms\n", tick + GetTickCount());
    fwrite(buffer, 1, count, stdout);
  }
  return 0;
}
